from typing import (
    Any,
    ClassVar,
    Generic,
    Literal,
    TypedDict,
    TypeVar,
)

from .base import DBModelMetaclass, RootModel


class InstanceId(RootModel):
    external_id: str
    space: str

    def __hash__(self) -> int:
        return hash((self.external_id, self.space))

    def __eq__(self, other: Any) -> bool:
        return (
            other is not None
            and isinstance(other, InstanceId)
            and self.external_id == other.external_id
            and self.space == other.space
        )

    def as_tuple(self) -> tuple[str, str]:
        return (self.space, self.external_id)


class ViewInstanceConfig(TypedDict, total=False):
    view_external_id: str | None
    instance_spaces: list[str] | None
    instance_spaces_prefix: str | None


class ViewInstance(InstanceId, metaclass=DBModelMetaclass):
    view_config: ClassVar[ViewInstanceConfig] = ViewInstanceConfig()

    @classmethod
    def get_view_external_id(cls) -> str:
        return cls.view_config.get("view_external_id") or cls.__name__


TViewInstance = TypeVar("TViewInstance", bound=ViewInstance)


class PaginatedResult(RootModel, Generic[TViewInstance]):
    data: list[TViewInstance]
    has_next_page: bool
    next_cursor: str | None


ValidationMode = Literal["raiseOnError", "ignoreOnError"]



from dataclasses import dataclass, field
from typing import Any, Generic, Self, TypeVar

from industrial_model.constants import DEFAULT_LIMIT, SORT_DIRECTION

from .expressions import (
    BoolExpression,
    Column,
    Expression,
    LeafExpression,
    and_,
    col,
    or_,
)

T = TypeVar("T")


@dataclass
class Statement(Generic[T]):
    entity: type[T] = field(init=True)
    where_clauses: list[Expression] = field(init=False, default_factory=list)
    sort_clauses: list[tuple[Column, SORT_DIRECTION]] = field(
        init=False, default_factory=list
    )
    limit_: int = field(init=False, default=DEFAULT_LIMIT)
    cursor_: str | None = field(init=False, default=None)

    def where(self, *expressions: bool | Expression) -> Self:
        for expression in expressions:
            assert isinstance(expression, Expression)
            self.where_clauses.append(expression)
        return self

    def asc(self, property: Any) -> Self:
        self.sort_clauses.append((Column(property), "ascending"))
        return self

    def desc(self, property: Any) -> Self:
        self.sort_clauses.append((Column(property), "descending"))
        return self

    def sort(self, property: Any, direction: SORT_DIRECTION) -> Self:
        self.sort_clauses.append((Column(property), direction))
        return self

    def limit(self, limit: int) -> Self:
        self.limit_ = limit
        return self

    def cursor(self, cursor: str | None) -> Self:
        self.cursor_ = cursor
        return self


def select(entity: type[T]) -> Statement[T]:
    return Statement(entity)


__all__ = [
    "Statement",
    "select",
    "Column",
    "col",
    "Expression",
    "LeafExpression",
    "BoolExpression",
    "and_",
    "or_",
]



from cognite.client.data_classes.data_modeling import InstanceSort, View

from industrial_model.cognite_adapters.utils import get_property_ref
from industrial_model.constants import SORT_DIRECTION
from industrial_model.statements.expressions import Column


class SortMapper:
    def map(
        self,
        sort_clauses: list[tuple[Column, SORT_DIRECTION]],
        root_view: View,
    ) -> list[InstanceSort]:
        return [
            InstanceSort(
                property=get_property_ref(column.property, root_view),
                direction=direction,
                nulls_first=direction == "descending",
            )
            for column, direction in sort_clauses
        ]



from industrial_model.models import RootModel, TViewInstance
from industrial_model.statements import Statement

from .params import NestedQueryParam, QueryParam, SortParam


class BaseQuery(RootModel):
    def to_statement(
        self, entity: type[TViewInstance]
    ) -> Statement[TViewInstance]:
        statement = Statement(entity)

        for key, item in self.__class__.model_fields.items():
            values = getattr(self, key)
            if not values:
                continue
            for metadata_item in item.metadata:
                if isinstance(metadata_item, SortParam):
                    statement.sort(values, metadata_item.direction)
                elif isinstance(metadata_item, QueryParam | NestedQueryParam):
                    statement.where(metadata_item.to_expression(values))

        return statement


class BasePaginatedQuery(BaseQuery):
    limit: int = 1000
    cursor: str | None = None

    def to_statement(
        self, entity: type[TViewInstance]
    ) -> Statement[TViewInstance]:
        statement = super().to_statement(entity)
        statement.limit(self.limit)
        statement.cursor(self.cursor)

        return statement



from industrial_model.models import RootModel


class DataModelId(RootModel):
    external_id: str
    space: str
    version: str

    def as_tuple(self) -> tuple[str, str, str]:
        return self.space, self.external_id, self.version



from cognite.client import CogniteClient

from industrial_model.models import TViewInstance
from industrial_model.statements import (
    BoolExpression,
    Expression,
    LeafExpression,
    Statement,
    col,
)

SPACE_PROPERTY = "space"


class QueryOptimizer:
    def __init__(
        self,
        cognite_client: CogniteClient,
    ):
        self._all_spaces: list[str] | None = None
        self._cognite_client = cognite_client

    def optimize(self, statement: Statement[TViewInstance]) -> None:
        instance_spaces = statement.entity.view_config.get("instance_spaces")
        instance_spaces_prefix = statement.entity.view_config.get(
            "instance_spaces_prefix"
        )

        if not instance_spaces and not instance_spaces_prefix:
            return

        if self._has_space_filter(statement.where_clauses):
            return

        filter_spaces = (
            self._find_spaces(instance_spaces_prefix)
            if instance_spaces_prefix
            else []
        )
        if instance_spaces:
            filter_spaces.extend(instance_spaces)

        if filter_spaces:
            statement.where(col(SPACE_PROPERTY).in_(filter_spaces))

    def _has_space_filter(self, where_clauses: list[Expression]) -> bool:
        for where_clause in where_clauses:
            if isinstance(
                where_clause, BoolExpression
            ) and self._has_space_filter(where_clause.filters):
                return True
            elif isinstance(where_clause, LeafExpression):
                return where_clause.property == SPACE_PROPERTY

        return False

    def _find_spaces(self, instance_spaces_prefix: str) -> list[str]:
        all_spaces = self._get_all_spaces()

        return [
            space
            for space in all_spaces
            if space.startswith(instance_spaces_prefix)
        ]

    def _get_all_spaces(self) -> list[str]:
        all_spaces = self._all_spaces
        if all_spaces is None:
            all_spaces = self._cognite_client.data_modeling.spaces.list(
                limit=-1
            ).as_ids()

        self._all_spaces = all_spaces
        return all_spaces



from typing import Any

from cognite.client import CogniteClient

from industrial_model.cognite_adapters.optimizer import QueryOptimizer
from industrial_model.config import DataModelId
from industrial_model.models import TViewInstance
from industrial_model.statements import Statement

from .query_mapper import QueryMapper
from .query_result_mapper import (
    QueryResultMapper,
)
from .view_mapper import ViewMapper


class CogniteAdapter:
    def __init__(
        self, cognite_client: CogniteClient, data_model_id: DataModelId
    ):
        self._cognite_client = cognite_client

        dm = cognite_client.data_modeling.data_models.retrieve(
            ids=data_model_id.as_tuple(),
            inline_views=True,
        ).latest_version()
        view_mapper = ViewMapper(dm.views)
        self._query_mapper = QueryMapper(view_mapper)
        self._result_mapper = QueryResultMapper(view_mapper)
        self._optmizer = QueryOptimizer(cognite_client)

    def query(
        self, statement: Statement[TViewInstance], all_pages: bool
    ) -> tuple[list[dict[str, Any]], str | None]:
        self._optmizer.optimize(statement)
        cognite_query = self._query_mapper.map(statement)
        view_external_id = statement.entity.get_view_external_id()

        data: list[dict[str, Any]] = []
        while True:
            query_result = self._cognite_client.data_modeling.instances.query(
                cognite_query
            )

            page_result, next_cursor = self._result_mapper.map_nodes(
                view_external_id, query_result
            )

            data.extend(page_result)

            last_page = len(page_result) < statement.limit_ or not next_cursor
            next_cursor_ = None if last_page else next_cursor
            cognite_query.cursors = {view_external_id: next_cursor_}

            if not all_pages or last_page:
                return data, next_cursor



from cognite.client import CogniteClient

from industrial_model.config import DataModelId
from industrial_model.models import (
    PaginatedResult,
    TViewInstance,
    ValidationMode,
)
from industrial_model.statements import Statement
from industrial_model.utils import run_async

from .engine import Engine


class AsyncEngine:
    def __init__(
        self,
        cognite_client: CogniteClient,
        data_model_id: DataModelId,
    ):
        self._engine = Engine(cognite_client, data_model_id)

    async def query_async(
        self,
        statement: Statement[TViewInstance],
        validation_mode: ValidationMode = "raiseOnError",
    ) -> PaginatedResult[TViewInstance]:
        return await run_async(self._engine.query, statement, validation_mode)

    async def query_all_pages_async(
        self,
        statement: Statement[TViewInstance],
        validation_mode: ValidationMode = "raiseOnError",
    ) -> list[TViewInstance]:
        return await run_async(
            self._engine.query_all_pages, statement, validation_mode
        )



import datetime
import json
from typing import Annotated

from pydantic import Field

from industrial_model import ViewInstance, ViewInstanceConfig, col, select

from .hubs import generate_engine


class DescribableEntity(ViewInstance):
    name: str | None = None
    description: str | None = None


class ReportingSite(DescribableEntity):
    time_zone: DescribableEntity | None = None


class EventDetail(ViewInstance):
    ref_metric_code: DescribableEntity | None = None
    ref_event_code: DescribableEntity | None = None
    ref_sub_category_l1: DescribableEntity | None = None
    ref_sub_category_l2: DescribableEntity | None = None
    ref_sub_category_l3: DescribableEntity | None = None
    ref_sub_category_l4: DescribableEntity | None = None
    ref_sub_category_l5: DescribableEntity | None = None
    ref_equipment: DescribableEntity | None = None
    ref_discipline: DescribableEntity | None = None


class Event(ViewInstance):
    view_config = ViewInstanceConfig(
        view_external_id="OEEEvent", instance_spaces_prefix="OEE-"
    )

    start_date_time: datetime.datetime | None = None
    ref_site: ReportingSite | None = None
    ref_unit: DescribableEntity | None = None
    ref_reporting_line: DescribableEntity | None = None
    ref_reporting_location: DescribableEntity | None = None
    ref_product: DescribableEntity | None = None
    ref_material: DescribableEntity | None = None
    ref_process_type: DescribableEntity | None = None
    ref_oee_event_detail: Annotated[
        list[EventDetail],
        Field(default_factory=list, alias="refOEEEventDetail"),
    ]


adapter = generate_engine()

filter = (
    col(Event.start_date_time).gt_(datetime.datetime(2025, 3, 1))
    & col(Event.ref_site).nested_(DescribableEntity.external_id == "STS-CLK")
    & (Event.start_date_time < datetime.datetime(2025, 6, 1))
)


statement = select(Event).limit(50).where(filter)

result = [
    item.model_dump(mode="json") for item in adapter.query_all_pages(statement)
]
print(len(result))
json.dump(
    result,
    open("test.json", "w"),
)



from datetime import datetime

import cognite.client.data_classes.filters as cdf_filters
from cognite.client.data_classes.data_modeling import MappedProperty, View

from industrial_model.cognite_adapters.utils import get_property_ref
from industrial_model.statements import (
    BoolExpression,
    Expression,
    LeafExpression,
)
from industrial_model.utils import datetime_to_ms_iso_timestamp

from .view_mapper import ViewMapper


class FilterMapper:
    def __init__(self, view_mapper: ViewMapper):
        self._view_mapper = view_mapper

    def map(
        self, expressions: list[Expression], root_view: View
    ) -> list[cdf_filters.Filter]:
        result: list[cdf_filters.Filter] = []
        for expression in expressions:
            if isinstance(expression, BoolExpression):
                result.append(self.to_cdf_filter_bool(expression, root_view))
            elif isinstance(expression, LeafExpression):
                result.append(self.to_cdf_filter_leaf(expression, root_view))
            else:
                cls_name = expression.__class__.__name__
                raise ValueError(f"Expression not implemented {cls_name}")
        return result

    def to_cdf_filter_bool(
        self, expression: BoolExpression, root_view: View
    ) -> cdf_filters.Filter:
        arguments = self.map(expression.filters, root_view)

        if expression.operator == "and":
            return cdf_filters.And(*arguments)
        elif expression.operator == "or":
            return cdf_filters.Or(*arguments)
        elif expression.operator == "not":
            return cdf_filters.Not(*arguments)

        raise NotImplementedError(f"Operator {self.operator} not implemented")

    def to_cdf_filter_leaf(
        self,
        expression: LeafExpression,
        root_view: View,
    ) -> cdf_filters.Filter:
        property_ref = get_property_ref(expression.property, root_view)

        value_ = expression.value
        if isinstance(value_, datetime):
            value_ = datetime_to_ms_iso_timestamp(value_)

        if expression.operator == "==":
            return cdf_filters.Equals(property_ref, value_)
        elif expression.operator == "in":
            return cdf_filters.In(property_ref, value_)
        elif expression.operator == ">":
            return cdf_filters.Range(property_ref, gt=value_)
        elif expression.operator == ">=":
            return cdf_filters.Range(property_ref, gte=value_)
        elif expression.operator == "<":
            return cdf_filters.Range(property_ref, lt=value_)
        elif expression.operator == "<=":
            return cdf_filters.Range(property_ref, lte=value_)
        elif expression.operator == "nested":
            target_view = self._get_nested_target_view(
                expression.property, root_view
            )

            assert isinstance(value_, Expression)

            return cdf_filters.Nested(
                property_ref,
                self.map([value_], target_view)[0],
            )
        elif expression.operator == "exists":
            return cdf_filters.Exists(property_ref)
        elif expression.operator == "prefix":
            return cdf_filters.Prefix(property_ref, value_)
        elif expression.operator == "containsAll":
            return cdf_filters.ContainsAll(property_ref, value_)
        elif expression.operator == "containsAny":
            return cdf_filters.ContainsAny(property_ref, value_)
        raise NotImplementedError(
            f"Operator {expression.operator} not implemented"
        )

    def _get_nested_target_view(self, property: str, root_view: View) -> View:
        view_definiton = root_view.properties[property]
        assert isinstance(view_definiton, MappedProperty)
        assert view_definiton.source
        return self._view_mapper.get_view(view_definiton.source.external_id)



from collections.abc import Callable, Generator
from datetime import datetime
from typing import (
    Any,
    ParamSpec,
    TypeVar,
)

from anyio import to_thread

TAny = TypeVar("TAny")
T_Retval = TypeVar("T_Retval")
P = ParamSpec("P")


def datetime_to_ms_iso_timestamp(dt: datetime) -> str:
    if not isinstance(dt, datetime):
        raise ValueError(f"Expected datetime object, got {type(dt)}")
    if dt.tzinfo is None:
        dt = dt.astimezone()
    return dt.isoformat(timespec="milliseconds")


def chunk_list(
    entries: list[TAny], chunk_size: int
) -> Generator[list[TAny], Any, None]:
    for i in range(0, len(entries), chunk_size):
        start = i
        end = i + chunk_size
        yield entries[start:end]


async def run_async(
    func: Callable[..., T_Retval],
    *args: object,
    cancellable: bool = False,
) -> T_Retval:
    return await to_thread.run_sync(func, *args, cancellable=cancellable)



from .base import RootModel
from .entities import (
    InstanceId,
    PaginatedResult,
    TViewInstance,
    ValidationMode,
    ViewInstance,
    ViewInstanceConfig,
)

__all__ = [
    "RootModel",
    "InstanceId",
    "TViewInstance",
    "ViewInstance",
    "ValidationMode",
    "PaginatedResult",
    "ViewInstanceConfig",
]



from collections import defaultdict
from typing import Any

from cognite.client.data_classes.data_modeling import (
    Edge,
    EdgeConnection,
    MappedProperty,
    Node,
    View,
)
from cognite.client.data_classes.data_modeling.query import (
    QueryResult,
)
from cognite.client.data_classes.data_modeling.views import (
    MultiReverseDirectRelation,
    SingleReverseDirectRelation,
)

from industrial_model.constants import EDGE_DIRECTION, EDGE_MARKER, NESTED_SEP

from .view_mapper import ViewMapper


class QueryResultMapper:
    def __init__(self, view_mapper: ViewMapper):
        self._view_mapper = view_mapper

    def map_nodes(
        self, root_node: str, query_result: QueryResult
    ) -> tuple[list[dict[str, Any]], str | None]:
        if root_node not in query_result:
            raise ValueError(
                f"{root_node} is not available in the query result"
            )

        root_view = self._view_mapper.get_view(root_node)

        values = self._map_node_property(root_node, root_view, query_result)
        if not values:
            return [], None

        data = [
            node
            for nodes in values.values()
            for node in self._nodes_to_dict(nodes)
        ]

        return data, query_result.cursors.get(root_node)

    def _map_node_property(
        self,
        key: str,
        view: View,
        query_result: QueryResult,
        result_property_key: str | None = None,
    ) -> dict[tuple[str, str], list[Node]] | None:
        if key not in query_result:
            return None

        mappings = self._get_property_mappings(key, view, query_result)

        view_id = view.as_id()

        nodes = (
            node
            for node in query_result.get_nodes(key)
            if isinstance(node, Node)
        )

        def get_node_id(node: Node) -> tuple[str, str]:
            if not result_property_key:
                return (node.space, node.external_id)

            entry = properties.get(result_property_key)
            if not isinstance(entry, dict):
                raise ValueError(
                    f"Invalid result property key {result_property_key}"
                )

            return entry.get("space", ""), entry.get("externalId", "")

        result: defaultdict[tuple[str, str], list[Node]] = defaultdict(list)
        for node in nodes:
            properties = node.properties.get(view_id, {})
            node_id = get_node_id(node)
            for mapping_key, (values, is_list) in mappings.items():
                element = properties.get(mapping_key)

                element_key: tuple[str, str] = (
                    (element.get("space", ""), element.get("externalId", ""))
                    if isinstance(element, dict)
                    else node_id
                )

                entries = values.get(element_key)
                if not entries:
                    continue

                entry_data = self._nodes_to_dict(entries)
                properties[mapping_key] = (
                    entry_data if is_list else entry_data[0]
                )

            node.properties[view_id] = properties

            result[node_id].append(node)

        return dict(result)

    def _get_property_mappings(
        self, key: str, view: View, query_result: QueryResult
    ) -> dict[str, tuple[dict[tuple[str, str], list[Node]], bool]]:
        mappings: dict[
            str, tuple[dict[tuple[str, str], list[Node]], bool]
        ] = {}

        for property_name, property in view.properties.items():
            property_key = f"{key}{NESTED_SEP}{property_name}"

            entry: dict[tuple[str, str], list[Node]] | None = None
            is_list = False

            if isinstance(property, MappedProperty) and property.source:
                entry = self._map_node_property(
                    property_key,
                    self._view_mapper.get_view(property.source.external_id),
                    query_result,
                )
                is_list = False
            elif (
                isinstance(property, SingleReverseDirectRelation)
                and property.source
            ):
                entry = self._map_node_property(
                    property_key,
                    self._view_mapper.get_view(property.source.external_id),
                    query_result,
                    property.through.property,
                )
                is_list = False
            elif (
                isinstance(property, MultiReverseDirectRelation)
                and property.source
            ):
                entry = self._map_node_property(
                    property_key,
                    self._view_mapper.get_view(property.source.external_id),
                    query_result,
                    property.through.property,
                )
                is_list = True

            elif isinstance(property, EdgeConnection) and property.source:
                entry = self._map_edge_property(
                    property_key,
                    self._view_mapper.get_view(property.source.external_id),
                    query_result,
                    property.direction,
                )
                is_list = True

            if entry:
                mappings[property_name] = entry, is_list

        return mappings

    def _map_edge_property(
        self,
        key: str,
        view: View,
        query_result: QueryResult,
        edge_direction: EDGE_DIRECTION,
    ) -> dict[tuple[str, str], list[Node]] | None:
        edge_key = f"{key}{NESTED_SEP}{EDGE_MARKER}"
        if key not in query_result or edge_key not in query_result:
            return None

        nodes = self._map_node_property(key, view, query_result)
        if not nodes:
            return None

        result: defaultdict[tuple[str, str], list[Node]] = defaultdict(list)

        edges = (
            edge
            for edge in query_result.get_edges(edge_key)
            if isinstance(edge, Edge)
        )
        for edge in edges:
            entry_key, node_key = (
                (
                    edge.end_node.as_tuple(),
                    edge.start_node.as_tuple(),
                )
                if edge_direction == "inwards"
                else (edge.start_node.as_tuple(), edge.end_node.as_tuple())
            )

            if node_item := nodes.get(node_key):
                result[entry_key].extend(node_item)

        return dict(result)

    def _nodes_to_dict(self, nodes: list[Node]) -> list[dict[str, Any]]:
        return [self._node_to_dict(node) for node in nodes]

    def _node_to_dict(self, node: Node) -> dict[str, Any]:
        entry = node.dump()
        properties: dict[str, dict[str, dict[str, Any]]] = (
            entry.pop("properties") or {}
        )
        for space_mapping in properties.values():
            for view_mapping in space_mapping.values():
                entry.update(view_mapping)

        return entry



from dataclasses import dataclass
from typing import Any

from industrial_model.constants import (
    LEAF_EXPRESSION_OPERATORS,
    SORT_DIRECTION,
)
from industrial_model.statements import LeafExpression


@dataclass
class QueryParam:
    property: str
    operator: LEAF_EXPRESSION_OPERATORS

    def to_expression(self, value: Any) -> LeafExpression:
        if self.operator == "nested":
            raise ValueError("Can not have nested operator on QuertParam")

        return LeafExpression(
            property=self.property,
            operator=self.operator,
            value=value,
        )


@dataclass
class NestedQueryParam:
    property: str
    value: QueryParam

    def to_expression(self, value: Any) -> LeafExpression:
        return LeafExpression(
            property=self.property,
            operator="nested",
            value=self.value.to_expression(value),
        )


@dataclass
class SortParam:
    direction: SORT_DIRECTION



from cognite.client.data_classes.data_modeling import (
    View,
)


class ViewMapper:
    def __init__(self, views: list[View]):
        self._views_as_dict = {view.external_id: view for view in views}

    def get_view(self, view_external_id: str) -> View:
        if view_external_id not in self._views_as_dict:
            raise ValueError(
                f"View {view_external_id} is not available in data model"
            )

        return self._views_as_dict[view_external_id]



from typing import Literal

from cognite.client.data_classes.data_modeling import (
    View,
    ViewId,
)

from industrial_model.models import TViewInstance

NODE_PROPERTIES = {"externalId", "space", "createdTime", "deletedTime"}
INTANCE_TYPE = Literal["node", "edge"]


def get_property_ref(
    property: str, view: View | ViewId, instance_type: INTANCE_TYPE = "node"
) -> tuple[str, str, str] | tuple[str, str]:
    return (
        view.as_property_ref(property)
        if property not in NODE_PROPERTIES
        else (instance_type, property)
    )


def get_cognite_instance_ids(
    instance_ids: list[TViewInstance],
) -> list[dict[str, str]]:
    return [
        get_cognite_instance_id(instance_id) for instance_id in instance_ids
    ]


def get_cognite_instance_id(instance_id: TViewInstance) -> dict[str, str]:
    return {"space": instance_id.space, "externalId": instance_id.external_id}



import os
from pathlib import Path
from string import Template
from typing import Any

import yaml
from cognite.client import CogniteClient
from dotenv import load_dotenv

from industrial_model import DataModelId, Engine


def generate_config(
    config_file_path: str | None = None,
) -> dict[str, dict[str, Any]]:
    load_dotenv(override=True)
    file_path = Path(
        config_file_path
        or f"{os.path.dirname(__file__)}/cognite-sdk-config.yaml"
    )
    env_sub_template = Template(file_path.read_text())
    file_env_parsed = env_sub_template.substitute(dict(os.environ))

    value = yaml.safe_load(file_env_parsed)

    assert isinstance(value, dict)
    return value


def generate_engine(config_file_path: str | None = None) -> Engine:
    cognite_config = generate_config(config_file_path)
    client = CogniteClient.load(cognite_config["cognite"])

    client.config.timeout = 60
    print(client.config)

    dm_id = DataModelId.model_validate(cognite_config["data_model"])
    return Engine(client, dm_id)



import cognite.client.data_classes.filters as filters
from cognite.client.data_classes.data_modeling import (
    EdgeConnection,
    MappedProperty,
    View,
    ViewId,
)
from cognite.client.data_classes.data_modeling.query import (
    EdgeResultSetExpression,
    NodeResultSetExpression,
    ResultSetExpression,
    Select,
    SourceSelector,
)
from cognite.client.data_classes.data_modeling.query import (
    Query as CogniteQuery,
)
from cognite.client.data_classes.data_modeling.views import (
    MultiReverseDirectRelation,
    SingleReverseDirectRelation,
)

from industrial_model.constants import EDGE_MARKER, MAX_LIMIT, NESTED_SEP
from industrial_model.models import TViewInstance
from industrial_model.statements import Statement

from .filter_mapper import (
    FilterMapper,
)
from .schemas import get_schema_properties
from .sort_mapper import SortMapper
from .view_mapper import ViewMapper


class QueryMapper:
    def __init__(self, view_mapper: ViewMapper):
        self._view_mapper = view_mapper
        self._filter_mapper = FilterMapper(view_mapper)
        self._sort_mapper = SortMapper()

    def map(self, statement: Statement[TViewInstance]) -> CogniteQuery:
        root_node = statement.entity.get_view_external_id()

        root_view = self._view_mapper.get_view(root_node)
        root_view_id = root_view.as_id()

        filters_: list[filters.Filter] = [
            filters.HasData(views=[root_view_id])
        ]

        filters_.extend(
            self._filter_mapper.map(statement.where_clauses, root_view)
        )

        with_: dict[str, ResultSetExpression] = {
            root_node: NodeResultSetExpression(
                filter=filters.And(*filters_),
                sort=self._sort_mapper.map(statement.sort_clauses, root_view),
                limit=statement.limit_,
            )
        }
        select_: dict[str, Select] = {}

        relations = get_schema_properties(
            statement.entity, NESTED_SEP, root_node
        )

        properties = self._include_statements(
            root_node, root_view, relations, with_, select_
        )

        select_[root_node] = self._get_select(root_view_id, properties)

        return CogniteQuery(
            with_=with_, select=select_, cursors={root_node: statement.cursor_}
        )

    def _get_select(self, view_id: ViewId, properties: list[str]) -> Select:
        return (
            Select(
                sources=[SourceSelector(source=view_id, properties=properties)]
            )
            if properties
            else Select()
        )

    def _include_statements(
        self,
        key: str,
        view: View,
        relations_to_include: list[str] | None,
        with_: dict[str, ResultSetExpression],
        select_: dict[str, Select],
    ) -> list[str]:
        if not relations_to_include:
            return []

        select_properties: list[str] = []
        for property_name, property in view.properties.items():
            property_key = f"{key}{NESTED_SEP}{property_name}"
            if property_key not in relations_to_include:
                continue

            if isinstance(property, MappedProperty) and not property.source:
                select_properties.append(property_name)
            elif isinstance(property, MappedProperty) and property.source:
                select_properties.append(property_name)

                props = self._include_statements(
                    property_key,
                    self._view_mapper.get_view(property.source.external_id),
                    relations_to_include,
                    with_,
                    select_,
                )
                if props:
                    with_[property_key] = NodeResultSetExpression(
                        from_=key,
                        through=view.as_property_ref(property_name),
                        limit=MAX_LIMIT,
                    )
                    select_[property_key] = self._get_select(
                        property.source, props
                    )

            elif (
                isinstance(property, MultiReverseDirectRelation)
                or isinstance(property, SingleReverseDirectRelation)
                and property.source
            ):
                props = self._include_statements(
                    property_key,
                    self._view_mapper.get_view(property.source.external_id),
                    relations_to_include,
                    with_,
                    select_,
                )
                if not props:
                    with_[property_key] = NodeResultSetExpression(
                        from_=key,
                        direction="inwards",
                        through=property.source.as_property_ref(
                            property.through.property
                        ),
                        limit=MAX_LIMIT,
                    )

                    select_[property_key] = self._get_select(
                        property.source, props
                    )
            elif isinstance(property, EdgeConnection) and property.source:
                edge_property_key = f"{property_key}{NESTED_SEP}{EDGE_MARKER}"

                with_[edge_property_key] = EdgeResultSetExpression(
                    from_=key,
                    max_distance=1,
                    filter=filters.Equals(
                        ["edge", "type"], property.type.dump()
                    ),
                    direction=property.direction,
                    limit=MAX_LIMIT,
                )
                with_[property_key] = NodeResultSetExpression(
                    from_=edge_property_key,
                    limit=MAX_LIMIT,
                )

                select_[edge_property_key] = Select()

                props = self._include_statements(
                    property_key,
                    self._view_mapper.get_view(property.source.external_id),
                    relations_to_include,
                    with_,
                    select_,
                )
                select_[property_key] = self._get_select(
                    property.source, props
                )

        return select_properties



from .models import BasePaginatedQuery, BaseQuery
from .params import NestedQueryParam, QueryParam, SortParam

__all__ = [
    "BaseQuery",
    "BasePaginatedQuery",
    "SortParam",
    "QueryParam",
    "NestedQueryParam",
]



from typing import Literal

SORT_DIRECTION = Literal["ascending", "descending"]
LEAF_EXPRESSION_OPERATORS = Literal[
    "==",
    "in",
    ">=",
    ">",
    "<=",
    "<",
    "nested",
    "exists",
    "prefix",
    "containsAll",
    "containsAny",
]
BOOL_EXPRESSION_OPERATORS = Literal["not", "and", "or"]


NESTED_SEP = "|"
EDGE_MARKER = "<EdgeMarker>"
EDGE_DIRECTION = Literal["outwards", "inwards"]
MAX_LIMIT = 10_000
DEFAULT_LIMIT = 5_000



import inspect
from collections import defaultdict
from types import UnionType
from typing import (
    Any,
    TypeVar,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)

from pydantic import BaseModel

from industrial_model.constants import NESTED_SEP

TBaseModel = TypeVar("TBaseModel", bound=BaseModel)


def get_schema_properties(
    cls: type[TBaseModel],
    nested_separator: str = NESTED_SEP,
    prefix: str | None = None,
) -> list[str]:
    data = _get_type_properties(cls) or {}
    keys = _flatten_dict_keys(data, None, nested_separator)
    if not prefix:
        return keys

    return [f"{prefix}{nested_separator}{key}" for key in keys]


def _get_type_properties(
    cls: type[BaseModel], visited_count: defaultdict[type, int] | None = None
) -> dict[str, Any] | None:
    if visited_count is not None:
        if visited_count[cls] > 2:
            return None
        visited_count[cls] += 1

    hints = get_type_hints(cls)
    origins = {
        key: _get_field_type(
            type_hint,
            visited_count or defaultdict(lambda: 0),
        )
        for key, type_hint in hints.items()
    }

    return {
        field_info.alias or key: origins[key][1]
        for key, field_info in cls.model_fields.items()
        if key in origins
    }


def _get_field_type(
    type_hint: type, visited_count: defaultdict[type, int]
) -> tuple[bool, dict[str, Any] | None]:
    should_iter = _type_is_list_or_union(type_hint)

    if not should_iter:
        return _get_field_relations(
            [_cast_base_model(type_hint)], visited_count
        )

    entries: list[type[BaseModel] | None] = []
    for arg in get_args(type_hint):
        if _type_is_list_or_union(arg):
            return _get_field_type(arg, visited_count)
        entries.append(_cast_base_model(arg))

    return _get_field_relations(entries, visited_count)


def _get_field_relations(
    entries: list[type[TBaseModel] | None],
    visited_count: defaultdict[type, int],
) -> tuple[bool, dict[str, Any] | None]:
    entry_type = next((type_ for type_ in entries if type_ is not None), None)

    if not entry_type:
        return False, None

    properties = _get_type_properties(entry_type, visited_count)

    return True, properties


def _type_is_list_or_union(entry: type) -> bool:
    origin = get_origin(entry)
    is_union = origin in (Union, UnionType)
    is_list = origin in (list, list)

    return is_union or is_list


def _cast_base_model(entry: type) -> type[TBaseModel] | None:
    is_base_model = (
        entry is not type(None)
        and inspect.isclass(entry) is True
        and issubclass(entry, BaseModel)
    )
    return entry if is_base_model else None


def _flatten_dict_keys(
    data: dict[str, Any], parent_key: str | None, nested_separator: str
) -> list[str]:
    paths: set[str] = set()
    for key, value in data.items():
        full_key = (
            f"{parent_key}{nested_separator}{key}" if parent_key else key
        )
        paths.add(full_key)
        if isinstance(value, dict) and value:
            paths.update(_flatten_dict_keys(value, full_key, nested_separator))
        elif isinstance(value, str):
            paths.add(f"{full_key}{nested_separator}{value}")
        elif isinstance(value, list | set):
            paths.update(
                [f"{full_key}{nested_separator}{item}" for item in value]
            )

    return sorted(paths)



from typing import Any

from cognite.client import CogniteClient

from industrial_model.cognite_adapters import CogniteAdapter
from industrial_model.config import DataModelId
from industrial_model.models import (
    PaginatedResult,
    TViewInstance,
    ValidationMode,
)
from industrial_model.statements import Statement


class Engine:
    def __init__(
        self,
        cognite_client: CogniteClient,
        data_model_id: DataModelId,
    ):
        self._cognite_adapter = CogniteAdapter(cognite_client, data_model_id)

    def query(
        self,
        statement: Statement[TViewInstance],
        validation_mode: ValidationMode = "raiseOnError",
    ) -> PaginatedResult[TViewInstance]:
        data, next_cursor = self._cognite_adapter.query(statement, False)

        return PaginatedResult(
            data=self._validate_data(statement.entity, data, validation_mode),
            next_cursor=next_cursor,
            has_next_page=next_cursor is not None,
        )

    def query_all_pages(
        self,
        statement: Statement[TViewInstance],
        validation_mode: ValidationMode = "raiseOnError",
    ) -> list[TViewInstance]:
        if statement.cursor_:
            raise ValueError("Cursor should be none when querying all pages")

        data, _ = self._cognite_adapter.query(statement, True)

        return self._validate_data(statement.entity, data, validation_mode)

    def _validate_data(
        self,
        entity: type[TViewInstance],
        data: list[dict[str, Any]],
        validation_mode: ValidationMode,
    ) -> list[TViewInstance]:
        result: list[TViewInstance] = []
        for item in data:
            try:
                result.append(entity.model_validate(item))
            except Exception:
                if validation_mode == "ignoreOnError":
                    continue
                raise
        return result



from .config import DataModelId
from .engines import AsyncEngine, Engine
from .models import (
    InstanceId,
    PaginatedResult,
    TViewInstance,
    ValidationMode,
    ViewInstance,
    ViewInstanceConfig,
)
from .statements import and_, col, or_, select

__all__ = [
    "and_",
    "or_",
    "col",
    "select",
    "ViewInstance",
    "InstanceId",
    "TViewInstance",
    "DataModelId",
    "ValidationMode",
    "Engine",
    "AsyncEngine",
    "PaginatedResult",
    "ViewInstanceConfig",
]
